<?php

namespace Drupal\gridstack;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Plugin\PluginBase;
use Drupal\gridstack\Entity\GridStack;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides base class for all gridstack plugins.
 */
abstract class GridStackPluginBase extends PluginBase implements GridStackPluginInterface {

  /**
   * The gridstack optionset.
   *
   * @var \Drupal\gridstack\Entity\GridStack
   */
  protected $optionset;

  /**
   * The layout breakpoints.
   *
   * @var array
   */
  protected $breakpoints = [];

  /**
   * The breakpoint columns.
   *
   * @var array
   */
  protected $columns = [];

  /**
   * The optionset min-width.
   *
   * @var int
   */
  protected $minWidth;

  /**
   * The optionset cell height.
   *
   * @var int
   */
  protected $cellHeight;

  /**
   * The optionset vertical margin.
   *
   * @var int
   */
  protected $margin;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * The gridstack manager service.
   *
   * @var \Drupal\gridstack\GridStackManagerInterface
   */
  protected $manager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = new static($configuration, $plugin_id, $plugin_definition);
    $instance->setConfiguration($configuration);

    if (!empty($configuration['optionset'])) {
      $instance->setOptionset($configuration['optionset']);
    }

    $instance->currentUser = $container->get('current_user');
    $instance->manager = $container->get('gridstack.manager');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function label() {
    return $this->pluginDefinition['label'];
  }

  /**
   * {@inheritdoc}
   */
  public function get($key) {
    return $this->pluginDefinition[$key] ?? NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $defaults = $this->configuration + $this->defaultConfiguration();
    $this->configuration = NestedArray::mergeDeep(
      $defaults,
      $configuration
    );
    return $this;
  }

  /**
   * Returns gridstack config shortcut.
   */
  protected function config($key = '', $settings = 'gridstack.settings') {
    return $this->manager->configLoad($key, $settings);
  }

  /**
   * {@inheritdoc}
   */
  public function getSetting($key, $default = NULL) {
    return $this->configuration[$key] ?? $default;
  }

  /**
   * {@inheritdoc}
   */
  public function setSetting($key, $value) {
    $this->configuration[$key] = $value;
    return $this;
  }

  /**
   * Sets the optionset.
   */
  public function setOptionset($optionset) {
    $this->optionset   = is_string($optionset) ? GridStack::loadWithFallback($optionset) : $optionset;
    $this->breakpoints = $this->optionset->breakpointsToArray();
    $this->columns     = $this->optionset->getColumns();
    $this->minWidth    = (int) $this->optionset->getSetting('minW', 481);
    $this->cellHeight  = $this->optionset->getSetting('cellHeight', 80);
    $this->margin      = $this->optionset->getSetting('margin', 0);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getOptionset() {
    if (!isset($this->optionset)) {
      $this->optionset = GridStack::loadWithFallback($this->getSetting('optionset', 'default'));
    }
    return $this->optionset;
  }

}
